/*
 *
 * Authors: Christos Bakogiannis (mpakoyan "at" gmail "dot" com)
 *          Nikolaos Karapanos   (karapas "at" gmail "dot" com)
 *
 *
 * This file is part of kmGNFS.
 *
 * This distribution of the General Number Field Sieve algorithm is
 * placed in the public domain by its authors. It is distributed with
 * the hope that it will be useful, and you can use it freely for any
 * purpose without having to notify anyone. However, if you find our
 * work useful in any way, please let us know about it.
 *
 * kmGNFS makes use of Victor Shoup's NTL free library.
 * NTL can be used in conjunction with GMP (the GNU Multi-Precision library)
 * for enhanced performance. For more info, visit NTL's home page.
 *
 * We disclaim any responsibility for any bug.
 *
 */


#include <iostream>
#include <cerrno>
#include <unistd.h>
#include <mpi.h>

#include <GNFS/gnfs_global.h>
#include <GNFS/Initialization.h>

#include <timer.h>

using namespace std;

int main(int argc, char* argv[]) {

    timer t;
    int my_id;
    int num_procs;
    MPI_Comm com;
    MPI_Status status;

    com = MPI_COMM_WORLD;

    //MPI::Init(argc, argv); //Initialize MPI.
    //num_procs = MPI::COMM_WORLD.Get_size(); //Get the number of processes.
    //my_id = MPI::COMM_WORLD.Get_rank(); //Determine the rank of this process.

    MPI_Init(&argc, &argv); //Initialize MPI.
    MPI_Comm_size(com, &num_procs); //Get the number of processes.
    MPI_Comm_rank(com, &my_id); //Determine the rank of this process.


    if (argc != 2) {
        if (my_id == MASTER)
            printGNFSUsage();

        return EXIT_FAILURE;
    }

    if (my_id == MASTER)
        cout << "*** Setting current working dir to \"" << argv[1] << "\" ..." << endl;


    errno = 0;
    if (chdir(argv[1]) == -1) {
        if (my_id == MASTER) {
            cout << "Error while trying to set current working directory... I am MPI process number " << my_id << endl;
            perror("The following error occured");
            printGNFSUsage();
        }

        return EXIT_FAILURE;
    }

    if (my_id == MASTER) {
#ifdef GNFS_DEBUG
        cout << "$$$ Note: GNFS debug mode is on... $$$" << endl;
#endif

        cout << "*** Initialization step started..." << endl;
        cout << "*** Reading input params..." << endl;
    }


    Initialization init;

    init.initMPI(com, status, num_procs, my_id); //todo

    if (my_id == MASTER) {
        cout << "*** Running some initial tests..." << endl;

        t.elapsed();
    }

    if (init.runInitialTests()) {
        if (my_id == MASTER)
            cout << "*** All tests passed..." << endl;
    }
    else {
        if (my_id == MASTER)
            cout << "*** Some tests failed... Aborting..." << endl;

        return EXIT_FAILURE;
    }


    if (init.performPolynomialSelection) {
        if (my_id == MASTER)
            cout << "*** Performing polynomial selection..." << endl;

        if (!init.polynomialSelectionStep()) {
            if (my_id == MASTER)
                cout << "An error occurred while trying to find the best rated polynomial through an extensive search..." << endl;

            return EXIT_FAILURE;
        }

        if (my_id == MASTER)
            cout << "*** Polynomial selection completed..." << endl;

    }
    else {
        if (my_id == MASTER)
            cout << "*** Skipping polynomial selection step..." << endl;
    }

    if (my_id == MASTER) {
        cout << "*** Writing output parameters file..." << endl;
        init.writeParamsFile();
        cout << endl << "*** All done! ***" << endl;

#ifdef GNFS_DEBUG
        cout << "$$$ Note: GNFS debug mode was on... $$$" << endl;
#endif

        double time = t.elapsed();
        cout << endl << "Total Time: " << time << endl;
    }

    //MPI::Finalize();
    MPI_Finalize();

    return EXIT_SUCCESS;
}
